from decimal import Decimal

from mm_solana import balance, token
from mm_std import Ok, Result

from mm_balance.constants import RETRIES_BALANCE, RETRIES_DECIMALS, TIMEOUT_BALANCE, TIMEOUT_DECIMALS


def get_native_balance(nodes: list[str], address: str, proxies: list[str], round_ndigits: int) -> Result[Decimal]:
    return balance.get_balance_with_retries(
        nodes, address, retries=RETRIES_BALANCE, timeout=TIMEOUT_BALANCE, proxies=proxies
    ).and_then(lambda b: Ok(round(Decimal(b / 1_000_000_000), round_ndigits)))


def get_token_balance(
    nodes: list[str], wallet_address: str, token_address: str, decimals: int, proxies: list[str], round_ndigits: int
) -> Result[Decimal]:
    return token.get_balance_with_retries(
        nodes, wallet_address, token_address, retries=RETRIES_BALANCE, timeout=TIMEOUT_BALANCE, proxies=proxies
    ).and_then(lambda b: Ok(round(Decimal(b / 10**decimals), round_ndigits)))


def get_token_decimals(nodes: list[str], token_address: str, proxies: list[str]) -> Result[int]:
    return token.get_decimals_with_retries(
        nodes, token_address, retries=RETRIES_DECIMALS, timeout=TIMEOUT_DECIMALS, proxies=proxies
    )



from __future__ import annotations

from decimal import Decimal

from mm_std import ConcurrentTasks, Result
from pydantic import BaseModel
from rich.progress import TaskID

from mm_balance import output
from mm_balance.config import Config
from mm_balance.constants import Network
from mm_balance.rpc import btc, eth, solana
from mm_balance.token_decimals import TokenDecimals


class Balances:
    class Balance(BaseModel):
        group_index: int
        address: str
        token_address: str | None
        balance: Result[Decimal] | None = None

    def __init__(self, config: Config, token_decimals: TokenDecimals):
        self.config = config
        self.token_decimals = token_decimals
        self.tasks: dict[Network, list[Balances.Balance]] = {network: [] for network in Network}
        self.progress_bar = output.create_progress_bar()
        self.progress_bar_task: dict[Network, TaskID] = {}

        for idx, group in enumerate(config.groups):
            task_list = [Balances.Balance(group_index=idx, address=a, token_address=group.token_address) for a in group.addresses]
            self.tasks[group.network].extend(task_list)

        for network in Network:
            if self.tasks[network]:
                self.progress_bar_task[network] = output.create_progress_task(
                    self.progress_bar, network.value, len(self.tasks[network])
                )

    def process(self) -> None:
        with self.progress_bar:
            job = ConcurrentTasks(max_workers=10)
            for network in Network:
                job.add_task(network.value, self._process_network, args=(network,))
            job.execute()

    def _process_network(self, network: Network) -> None:
        job = ConcurrentTasks(max_workers=self.config.workers[network])
        for idx, task in enumerate(self.tasks[network]):
            job.add_task(str(idx), self._get_balance, args=(network, task.address, task.token_address))
        job.execute()
        for idx, _task in enumerate(self.tasks[network]):
            self.tasks[network][idx].balance = job.result.get(str(idx))  # type: ignore[assignment]

    def _get_balance(self, network: Network, wallet_address: str, token_address: str | None) -> Result[Decimal]:
        nodes = self.config.nodes[network]
        round_ndigits = self.config.round_ndigits
        proxies = self.config.proxies
        token_decimals = self.token_decimals[network][token_address] if token_address else -1
        match network:
            case Network.BITCOIN:
                res = btc.get_balance(wallet_address, proxies, round_ndigits)
            case Network.ETHEREUM | Network.ARBITRUM_ONE | Network.OP_MAINNET:
                if token_address is None:
                    res = eth.get_native_balance(nodes, wallet_address, proxies, round_ndigits)
                else:
                    res = eth.get_token_balance(nodes, wallet_address, token_address, token_decimals, proxies, round_ndigits)
            case Network.SOLANA:
                if token_address is None:
                    res = solana.get_native_balance(nodes, wallet_address, proxies, round_ndigits)
                else:
                    res = solana.get_token_balance(nodes, wallet_address, token_address, token_decimals, proxies, round_ndigits)

            case _:
                raise ValueError

        self.progress_bar.update(self.progress_bar_task[network], advance=1)
        return res

    def get_group_balances(self, group_index: int, network: Network) -> list[Balance]:
        # TODO: can we get network by group_index?
        return [b for b in self.tasks[network] if b.group_index == group_index]

    def get_errors(self) -> list[Balance]:
        result = []
        for network in self.tasks:
            for task in self.tasks[network]:
                if task.balance is not None and task.balance.is_err():
                    result.append(task)
        return result



from __future__ import annotations

from enum import Enum, unique

RETRIES_BALANCE = 5
RETRIES_DECIMALS = 5
RETRIES_COINGECKO_PRICES = 5
TIMEOUT_BALANCE = 5
TIMEOUT_DECIMALS = 5


@unique
class Network(str, Enum):
    ARBITRUM_ONE = "arbitrum-one"
    BITCOIN = "bitcoin"
    ETHEREUM = "ethereum"
    SOLANA = "solana"
    OP_MAINNET = "op-mainnet"  # Optimism mainnet


TOKEN_ADDRESS: dict[Network, dict[str, str]] = {
    Network.ETHEREUM: {
        "USDT": "0xdac17f958d2ee523a2206206994597c13d831ec7",
        "USDC": "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48",
    },
    Network.SOLANA: {
        "USDT": "Es9vMFrzaCERmJfrF4H2FYD4KCoNkY11McCe8BenwNYB",
        "USDC": "EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v",
    },
    Network.ARBITRUM_ONE: {
        "USDT": "0xfd086bc7cd5c481dcc9c85ebe478a1c0b69fcbb9",
        "USDC": "0xff970a61a04b1ca14834a43f5de4533ebddb5cc8",
    },
    Network.OP_MAINNET: {
        "USDT": "0x94b008aA00579c1307B0EF2c499aD98a8ce58e58",
        "USDC": "0x7f5c764cbc14f9669b88837ca1490cca17c31607",
    },
}

TICKER_TO_COINGECKO_ID = {
    "BTC": "bitcoin",
    "ETH": "ethereum",
    "USDT": "tether",
    "USDC": "usd-coin",
    "SOL": "solana",
}

USD_STABLECOINS = ["USDT", "USDC"]

DEFAULT_NODES: dict[Network, list[str]] = {
    Network.ARBITRUM_ONE: ["https://arb1.arbitrum.io/rpc", "https://arbitrum.llamarpc.com"],
    Network.BITCOIN: [],
    Network.ETHEREUM: ["https://ethereum.publicnode.com", "https://rpc.ankr.com/eth"],
    Network.SOLANA: ["https://api.mainnet-beta.solana.com"],
    Network.OP_MAINNET: ["https://mainnet.optimism.io", "https://optimism.llamarpc.com"],
}



from decimal import Decimal

from mm_eth import erc20, rpc
from mm_std import Ok, Result

from mm_balance.constants import RETRIES_BALANCE, RETRIES_DECIMALS, TIMEOUT_BALANCE, TIMEOUT_DECIMALS


def get_native_balance(nodes: list[str], address: str, proxies: list[str], round_ndigits: int) -> Result[Decimal]:
    return rpc.eth_get_balance(nodes, address, proxies=proxies, attempts=RETRIES_BALANCE, timeout=TIMEOUT_BALANCE).and_then(
        lambda b: Ok(round(Decimal(b / 10**18), round_ndigits)),
    )


def get_token_balance(
    nodes: list[str], wallet_address: str, token_address: str, decimals: int, proxies: list[str], round_ndigits: int
) -> Result[Decimal]:
    return erc20.get_balance(
        nodes,
        token_address,
        wallet_address,
        proxies=proxies,
        attempts=RETRIES_BALANCE,
        timeout=TIMEOUT_BALANCE,
    ).and_then(
        lambda b: Ok(round(Decimal(b / 10**decimals), round_ndigits)),
    )


def get_token_decimals(nodes: list[str], token_address: str, proxies: list[str]) -> Result[int]:
    return erc20.get_decimals(nodes, token_address, timeout=TIMEOUT_DECIMALS, proxies=proxies, attempts=RETRIES_DECIMALS)



import getpass
import pathlib
import pkgutil
from typing import Annotated

import typer

from mm_balance import output
from mm_balance.balances import Balances
from mm_balance.config import Config
from mm_balance.constants import Network
from mm_balance.price import Prices, get_prices
from mm_balance.token_decimals import get_token_decimals

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False, add_completion=False)


def example_callback(value: bool) -> None:
    if value:
        data = pkgutil.get_data(__name__, "config/example.yml")
        typer.echo(data)
        raise typer.Exit


def networks_callback(value: bool) -> None:
    if value:
        for network in Network:
            typer.echo(network)
        raise typer.Exit


@app.command()
def cli(
    config_path: Annotated[pathlib.Path, typer.Argument()],
    _example: Annotated[bool | None, typer.Option("--example", callback=example_callback, help="Print a config example.")] = None,
    _networks: Annotated[
        bool | None, typer.Option("--networks", callback=networks_callback, help="Print supported networks.")
    ] = None,
) -> None:
    zip_password = ""  # nosec
    if config_path.name.endswith(".zip"):
        zip_password = getpass.getpass("zip password")
    config = Config.read_config(config_path, zip_password=zip_password)

    prices = get_prices(config) if config.price else Prices()
    output.print_prices(config, prices)

    token_decimals = get_token_decimals(config)
    balances = Balances(config, token_decimals)
    balances.process()

    output.print_groups(balances, config, prices)
    output.print_total(config, balances, prices)
    output.print_errors(config, balances)


if __name__ == "__main__":
    app()



from __future__ import annotations

from decimal import Decimal
from pathlib import Path
from typing import Self

import pydash
from mm_std import BaseConfig, PrintFormat, fatal, hr
from pydantic import Field, field_validator, model_validator

from mm_balance.constants import DEFAULT_NODES, TOKEN_ADDRESS, Network


class Group(BaseConfig):
    comment: str = ""
    ticker: str
    network: Network
    token_address: str | None = None
    coingecko_id: str | None = None
    addresses: list[str] = Field(default_factory=list)
    share: Decimal = Decimal(1)

    @property
    def name(self) -> str:
        result = self.ticker
        if self.comment:
            result += " / " + self.comment
        result += " / " + self.network.value
        return result

    @field_validator("ticker", mode="after")
    def ticker_validator(cls, v: str) -> str:
        return v.upper()

    @field_validator("addresses", mode="before")
    def to_list_validator(cls, v: str | list[str] | None) -> list[str]:
        return cls.to_list_str_validator(v, unique=True, remove_comments=True, split_line=True)

    # @model_validator(mode="before")
    # def before_all(cls, data: Any) -> Any:
    #     if "network" not in data:
    #         data["network"] = detect_network(data["coin"])
    #     return data

    @model_validator(mode="after")
    def final_validator(self) -> Self:
        if self.token_address is None:
            self.token_address = detect_token_address(self.ticker, self.network)
        if self.token_address is not None and self.network is Network.ETHEREUM:
            self.token_address = self.token_address.lower()
        return self

    def process_addresses(self, address_groups: list[AddressGroup]) -> None:
        addresses: list[str] = []
        for address in self.addresses:
            if address_group := pydash.find(address_groups, lambda g: g.name == address):  # noqa: B023
                addresses.extend(address_group.addresses)
            else:
                # TODO: check address is valid
                addresses.append(address)
        self.addresses = pydash.uniq(process_file_addresses(addresses))


class AddressGroup(BaseConfig):
    name: str
    addresses: list[str]

    @field_validator("addresses", mode="before")
    def to_list_validator(cls, v: str | list[str] | None) -> list[str]:
        return cls.to_list_str_validator(v, unique=True, remove_comments=True, split_line=True)


class Config(BaseConfig):
    groups: list[Group] = Field(alias="coins")
    addresses: list[AddressGroup] = Field(default_factory=list)

    proxies_url: str | None = None
    proxies: list[str] = Field(default_factory=list)
    round_ndigits: int = 4
    nodes: dict[Network, list[str]] = Field(default_factory=dict)
    print_format: PrintFormat = PrintFormat.TABLE
    price: bool = True
    skip_empty: bool = False  # don't print the address with an empty balance

    workers: dict[Network, int] = {network: 5 for network in Network}

    def has_share(self) -> bool:
        return any(g.share != Decimal(1) for g in self.groups)

    @model_validator(mode="after")
    def final_validator(self) -> Self:
        # load from proxies_url
        if self.proxies_url is not None:
            self.proxies = get_proxies(self.proxies_url)

        # load addresses from address_group
        for group in self.groups:
            group.process_addresses(self.addresses)

        # load default rpc nodes
        for network in Network:
            if network not in self.nodes:
                self.nodes[network] = DEFAULT_NODES[network]

        return self


def detect_token_address(ticker: str, network: Network) -> str | None:
    if network in TOKEN_ADDRESS:
        return TOKEN_ADDRESS[network].get(ticker)


def get_proxies(proxies_url: str) -> list[str]:
    try:
        res = hr(proxies_url)
        if res.is_error():
            fatal(f"Can't get proxies: {res.error}")
        proxies = [p.strip() for p in res.body.splitlines() if p.strip()]
        return pydash.uniq(proxies)
    except Exception as err:
        fatal(f"Can't get  proxies: {err}")


def get_address_group_by_name(address_groups: list[AddressGroup], name: str) -> AddressGroup | None:
    return pydash.find(address_groups, lambda g: g.name == name)


def process_file_addresses(addresses: list[str]) -> list[str]:
    result = []
    for address in addresses:
        if address.startswith("file://"):
            path = Path(address.removeprefix("file://"))
            if path.is_file():
                result.extend(path.read_text().strip().splitlines())
            else:
                fatal(f"File with addresses not found: {path}")
        else:
            result.append(address)
    return result



from decimal import Decimal

from mm_std import Ok, print_table
from rich.progress import BarColumn, MofNCompleteColumn, Progress, TaskID, TextColumn

from mm_balance.balances import Balances
from mm_balance.config import Config, Group
from mm_balance.price import Prices
from mm_balance.total import Total


def print_groups(balances: Balances, config: Config, prices: Prices) -> None:
    for group_index, group in enumerate(config.groups):
        group_balances = balances.get_group_balances(group_index, group.network)
        _print_group(group, group_balances, config, prices)


def _print_group(group: Group, group_balances: list[Balances.Balance], config: Config, prices: Prices) -> None:
    rows = []
    balance_sum = Decimal(0)
    usd_sum = Decimal(0)
    for address_task in group_balances:
        if config.skip_empty and isinstance(address_task.balance, Ok) and address_task.balance.ok == Decimal(0):
            continue
        row = [address_task.address, address_task.balance.ok_or_err()]  # type: ignore[union-attr]
        if isinstance(address_task.balance, Ok):
            balance_sum += address_task.balance.ok
            if config.price:
                balance_usd = round(address_task.balance.ok * prices[group.ticker], config.round_ndigits)
                usd_sum += balance_usd
                row.append(f"${balance_usd}")
        rows.append(row)

    sum_row = ["sum", round(balance_sum, config.round_ndigits)]
    if config.price:
        sum_row.append(f"${round(usd_sum, config.round_ndigits)}")
    rows.append(sum_row)

    if group.share < Decimal(1):
        sum_share_row = [f"sum_share, {group.share}", round(balance_sum * group.share, config.round_ndigits)]
        if config.price:
            sum_share_row.append(f"${round(usd_sum * group.share, config.round_ndigits)}")
        rows.append(sum_share_row)

    table_headers = ["address", "balance"]
    if config.price:
        table_headers += ["usd"]
    print_table(group.name, table_headers, rows)


def print_prices(config: Config, prices: Prices) -> None:
    if config.price:
        rows = [[k, round(v, config.round_ndigits)] for (k, v) in prices.items()]
        print_table("Prices", ["coin", "usd"], rows)


def print_total(config: Config, balances: Balances, prices: Prices) -> None:
    total = Total.calc(balances, prices, config)
    total.print()


def print_errors(config: Config, balances: Balances) -> None:
    error_balances = balances.get_errors()
    if not error_balances:
        return
    rows = []
    for balance in error_balances:
        group = config.groups[balance.group_index]
        rows.append([group.ticker + " / " + group.network, balance.address, balance.balance.err])  # type: ignore[union-attr]
    print_table("Errors", ["coin", "address", "error"], rows)


def create_progress_bar() -> Progress:
    return Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        MofNCompleteColumn(),
    )


def create_progress_task(progress: Progress, description: str, total: int) -> TaskID:
    return progress.add_task("[green]" + description, total=total)



from collections import defaultdict
from dataclasses import dataclass
from decimal import Decimal
from typing import Self

from mm_std import Ok, PrintFormat, print_table

from mm_balance.balances import Balances
from mm_balance.config import Config
from mm_balance.constants import USD_STABLECOINS
from mm_balance.price import Prices


@dataclass
class Total:
    coins: dict[str, Decimal]
    coins_share: dict[str, Decimal]
    usd_sum: Decimal  # sum of all coins in USD
    usd_sum_share: Decimal

    stablecoin_sum: Decimal  # sum of usd stablecoins: usdt, usdc
    stablecoin_sum_share: Decimal

    config: Config
    prices: Prices

    @classmethod
    def calc(cls, balances: Balances, prices: Prices, config: Config) -> Self:
        coins: dict[str, Decimal] = defaultdict(Decimal)
        coins_share: dict[str, Decimal] = defaultdict(Decimal)
        usd_sum = Decimal(0)
        usd_sum_share = Decimal(0)

        stablecoin_sum = Decimal(0)
        stablecoin_sum_share = Decimal(0)
        for group_index, group in enumerate(config.groups):
            balance_sum = Decimal(0)
            for address_task in balances.get_group_balances(group_index, group.network):
                if isinstance(address_task.balance, Ok):
                    balance_sum += address_task.balance.ok
                    if group.ticker in USD_STABLECOINS:
                        stablecoin_sum += address_task.balance.ok
                        stablecoin_sum_share += address_task.balance.ok * group.share
                    if config.price:
                        balance_usd = round(address_task.balance.ok * prices[group.ticker], config.round_ndigits)
                        usd_sum += balance_usd
                        usd_sum_share += group.share * balance_usd

            coins[group.ticker] += balance_sum
            coins_share[group.ticker] += round(balance_sum * group.share, config.round_ndigits)
        return cls(
            coins=coins,
            coins_share=coins_share,
            usd_sum=usd_sum,
            usd_sum_share=usd_sum_share,
            # usd_share=usd_share,
            stablecoin_sum=stablecoin_sum,
            stablecoin_sum_share=stablecoin_sum_share,
            config=config,
            prices=prices,
        )

    def print(self) -> None:
        if self.config.print_format == PrintFormat.TABLE:
            if self.config.price:
                self._print_total_total_with_price()

                if self.config.has_share():
                    self._print_share_total_with_price()
            else:
                self._print_total_total_without_price()

                if self.config.has_share():
                    self._print_share_total_without_price()

    def _print_total_total_with_price(self) -> None:
        if self.config.print_format == PrintFormat.TABLE:
            rows = []
            for key, value in self.coins.items():
                usd_value = round(value * self.prices[key], self.config.round_ndigits)
                if key in USD_STABLECOINS:
                    usd_share = round(self.stablecoin_sum * 100 / self.usd_sum, self.config.round_ndigits)
                else:
                    usd_share = round(usd_value * 100 / self.usd_sum, self.config.round_ndigits)
                rows.append([key, value, f"${usd_value}", f"{usd_share}%"])
            rows.append(["usd_sum", f"${self.usd_sum}"])
            print_table("Total", ["coin", "balance", "usd", "usd_share"], rows)

    def _print_total_total_without_price(self) -> None:
        if self.config.print_format == PrintFormat.TABLE:
            rows = []
            for key, value in self.coins.items():
                rows.append([key, value])
            print_table("Total", ["coin", "balance"], rows)

    def _print_share_total_with_price(self) -> None:
        rows = []
        for key, _ in self.coins.items():
            usd_value = round(self.coins_share[key] * self.prices[key], self.config.round_ndigits)
            if key in USD_STABLECOINS:
                usd_share = round(self.stablecoin_sum_share * 100 / self.usd_sum_share, self.config.round_ndigits)
            else:
                usd_share = round(usd_value * 100 / self.usd_sum_share, self.config.round_ndigits)
            rows.append([key, self.coins_share[key], f"${usd_value}", f"{usd_share}%"])
        rows.append(["usd_sum", f"${self.usd_sum_share}"])
        print_table("Total, share", ["coin", "balance", "usd", "usd_share"], rows)

    def _print_share_total_without_price(self) -> None:
        rows = []
        for key, _ in self.coins.items():
            rows.append([key, self.coins_share[key]])
        print_table("Total, share", ["coin", "balance"], rows)



from mm_std import Err, fatal

from mm_balance.config import Config
from mm_balance.constants import Network
from mm_balance.rpc import eth, solana


class TokenDecimals(dict[Network, dict[str, int]]):
    def __init__(self) -> None:
        super().__init__()
        for network in Network:
            self[network] = {}


def get_token_decimals(config: Config) -> TokenDecimals:
    result = TokenDecimals()

    for group in config.groups:
        if group.token_address is None or group.token_address in result[group.network]:
            continue

        nodes = config.nodes[group.network]
        proxies = config.proxies

        match group.network:
            case Network.ETHEREUM:
                decimals_res = eth.get_token_decimals(nodes, group.token_address, proxies)
            case Network.SOLANA:
                decimals_res = solana.get_token_decimals(nodes, group.token_address, proxies)
            case _:
                raise ValueError(f"unsupported network: {group.network}. Cant get token decimals for {group.token_address}")

        if isinstance(decimals_res, Err):
            fatal(f"can't get decimals for token {group.ticker} / {group.token_address}, error={decimals_res.err}")
        result[group.network][group.token_address] = decimals_res.ok

    return result



from decimal import Decimal

from mm_btc.blockstream import BlockstreamClient
from mm_std import Ok, Result

from mm_balance.constants import RETRIES_BALANCE


def get_balance(address: str, proxies: list[str], round_ndigits: int) -> Result[Decimal]:
    return (
        BlockstreamClient(proxies=proxies, attempts=RETRIES_BALANCE)
        .get_confirmed_balance(address)
        .and_then(
            lambda b: Ok(round(Decimal(b / 100_000_000), round_ndigits)),
        )
    )



from decimal import Decimal

import pydash
from mm_std import fatal, hr
from mm_std.random_ import random_str_choice

from mm_balance.config import Config, Group
from mm_balance.constants import RETRIES_COINGECKO_PRICES, Network


class Prices(dict[str, Decimal]):
    """
    A Prices class representing a mapping from coin names to their prices.

    Inherits from:
        Dict[str, Decimal]: A dictionary with coin names as keys and their prices as Decimal values.
    """


def get_prices(config: Config) -> Prices:
    result = Prices()

    coins = pydash.uniq([group.ticker for group in config.groups])
    coingecko_ids = pydash.uniq([get_coingecko_id(group) for group in config.groups])

    url = f"https://api.coingecko.com/api/v3/simple/price?ids={",".join(coingecko_ids)}&vs_currencies=usd"
    for _ in range(RETRIES_COINGECKO_PRICES):
        res = hr(url, proxy=random_str_choice(config.proxies))
        if res.code != 200:
            continue

        for idx, coin in enumerate(coins):
            if coingecko_ids[idx] in res.json:
                result[coin] = Decimal(str(pydash.get(res.json, f"{coingecko_ids[idx]}.usd")))
            else:
                fatal("Can't get price for {coin} from coingecko, coingecko_id={coingecko_ids[idx]}")

    return result


TICKER_TO_COINGECKO_ID = {
    "BTC": "bitcoin",
    "ETH": "ethereum",
    "USDT": "tether",
    "USDC": "usd-coin",
    "SOL": "solana",
}


def get_coingecko_id(group: Group) -> str:
    if group.coingecko_id:
        return group.coingecko_id
    elif group.network is Network.BITCOIN:
        return "bitcoin"
    elif group.network is Network.ETHEREUM and group.token_address is None:
        return "ethereum"
    elif group.ticker == "ETH":
        return "ethereum"
    elif group.ticker == "USDT":
        return "tether"
    elif group.ticker == "USDC":
        return "usd-coin"
    elif group.ticker == "SOL":
        return "solana"

    raise ValueError(f"can't get coingecko_id for {group.ticker}")
